@@ -64,7 +64,7 @@ atk_hyperlink_get_type (void)
 static void
 atk_hyperlink_class_init (AtkHyperlinkClass *klass)
 {
-  parent_class = g_type_class_ref (G_TYPE_OBJECT);
+  parent_class = g_type_class_peek_parent (klass);
 
 }
 
