@@ -19,44 +19,19 @@
 
 #include "atktext.h"
 #include "atkmarshal.h"
+#include "atk-enum-types.h"
+
+#include <string.h>
+
+GPtrArray *extra_attributes = NULL;
 
 enum {
   TEXT_CHANGED,
-  CARET_MOVED,
+  TEXT_CARET_MOVED,
   TEXT_SELECTION_CHANGED,
   LAST_SIGNAL
 };
 
-static const gchar * text_attr_name[] = {
-  "left_margin",
-  "right_margin",
-  "indent",
-  "invisible",
-  "editable",
-  "pixels_above_lines",
-  "pixels_below_lines",
-  "pixels_inside_wrap",
-  "bg_full_height",
-  "rise",
-  "underline",
-  "strikethrough",
-  "size",
-  "scale",
-  "weight",
-  "language",
-  "family_name",
-  "bg_color",
-  "fg_color",
-  "bg_stipple",
-  "fg_stipple",
-  "wrap_mode",
-  "direction",
-  "justification",
-  "stretch",
-  "variant",
-  "slant_style",
-};
-
 static const gchar *bool[] = {"false",
                               "true"};
 static const gchar *style[] = {"normal",
@@ -76,7 +51,6 @@ static const gchar *stretch[] = {"ultra_condensed",
 static const gchar *justification[] = {"left",
                                        "right",
                                        "center",
-
                                        "fill"};
 static const gchar *direction[] = {"none",
                                    "ltr",
@@ -138,11 +112,11 @@ atk_text_base_init (gpointer *g_class)
 		      G_TYPE_NONE,
 		      2, G_TYPE_INT, G_TYPE_INT);
       
-      atk_text_signals[CARET_MOVED] =
+      atk_text_signals[TEXT_CARET_MOVED] =
 	g_signal_new ("text_caret_moved",
 		      ATK_TYPE_TEXT,
 		      G_SIGNAL_RUN_LAST,
-		      G_STRUCT_OFFSET (AtkTextIface, caret_changed),
+		      G_STRUCT_OFFSET (AtkTextIface, text_caret_moved),
 		      (GSignalAccumulator) NULL, NULL,
 		      g_cclosure_marshal_VOID__INT,
 		      G_TYPE_NONE,
@@ -182,6 +156,9 @@ atk_text_get_text (AtkText      *text,
 
   iface = ATK_TEXT_GET_IFACE (text);
 
+  if (start_offset < 0 || end_offset < 0)
+    return NULL;
+
   if (iface->get_text)
     return (*(iface->get_text)) (text, start_offset, end_offset);
   else
@@ -205,6 +182,9 @@ atk_text_get_character_at_offset (AtkText      *text,
 
   g_return_val_if_fail (ATK_IS_TEXT (text), (gunichar) 0);
 
+  if (offset < 0)
+    return (gunichar) 0;
+
   iface = ATK_TEXT_GET_IFACE (text);
 
   if (iface->get_character_at_offset)
@@ -283,6 +263,9 @@ atk_text_get_text_after_offset (AtkText          *text,
   else
     real_end_offset = &local_end_offset;
 
+  if (offset < 0)
+    return NULL;
+
   iface = ATK_TEXT_GET_IFACE (text);
 
   if (iface->get_text_after_offset)
@@ -368,6 +351,9 @@ atk_text_get_text_at_offset (AtkText          *text,
   else
     real_end_offset = &local_end_offset;
 
+  if (offset < 0)
+    return NULL;
+
   iface = ATK_TEXT_GET_IFACE (text);
 
   if (iface->get_text_at_offset)
@@ -451,6 +437,9 @@ atk_text_get_text_before_offset (AtkText          *text,
   else
     real_end_offset = &local_end_offset;
 
+  if (offset < 0)
+    return NULL;
+
   iface = ATK_TEXT_GET_IFACE (text);
 
   if (iface->get_text_before_offset)
@@ -527,17 +516,18 @@ atk_text_get_character_extents (AtkText *text,
   else
     real_height = &local_height;
 
+  *real_x = 0;
+  *real_y = 0;
+  *real_width = 0;
+  *real_height = 0;
+
+  if (offset < 0)
+    return;
+ 
   iface = ATK_TEXT_GET_IFACE (text);
 
   if (iface->get_character_extents)
     (*(iface->get_character_extents)) (text, offset, real_x, real_y, real_width, real_height, coords);
-  else
-    {
-      *real_x = 0;
-      *real_y = 0;
-      *real_width = 0;
-      *real_height = 0;
-    }
 }
 
 /**
@@ -578,6 +568,9 @@ atk_text_get_run_attributes (AtkText          *text,
   else
     real_end_offset = &local_end_offset;
 
+  if (offset < 0)
+    return NULL;
+
   iface = ATK_TEXT_GET_IFACE (text);
 
   if (iface->get_run_attributes)
@@ -870,7 +863,7 @@ atk_text_set_caret_offset (AtkText *text,
  * #AtkAttributes.
  **/
 void
-atk_attribute_set_free(AtkAttributeSet *attrib_set)
+atk_attribute_set_free (AtkAttributeSet *attrib_set)
 {
   GSList *temp;
 
@@ -891,7 +884,27 @@ atk_attribute_set_free(AtkAttributeSet *attrib_set)
 }
 
 /**
- * atk_attribute_get_name:
+ * atk_text_attribute_register:
+ * @name: a name string
+ *
+ * Associate @name with a new #AtkTextAttribute
+ *
+ * Returns: an #AtkTextAttribute associated with @name
+ **/
+AtkTextAttribute
+atk_text_attribute_register (const gchar *name)
+{
+  g_return_val_if_fail (name, ATK_TEXT_ATTR_INVALID);
+
+  if (!extra_attributes)
+    extra_attributes = g_ptr_array_new ();
+
+  g_ptr_array_add (extra_attributes, g_strdup (name));
+  return extra_attributes->len + ATK_TEXT_ATTR_LAST_DEFINED;
+}
+
+/**
+ * atk_text_attribute_get_name:
  * @attr: The #AtkTextAttribute whose name is required
  *
  * Gets the name corresponding to the #AtkTextAttribute
@@ -899,14 +912,94 @@ atk_attribute_set_free(AtkAttributeSet *attrib_set)
  * Returns: a string containing the name; this string should not be freed
  **/
 G_CONST_RETURN gchar*
-atk_attribute_get_name (AtkTextAttribute attr)
+atk_text_attribute_get_name (AtkTextAttribute attr)
 {
-  g_assert (attr >= 0 && attr <= ATK_TEXT_ATTR_STYLE);
-  return text_attr_name[attr];
+  GTypeClass *type_class;
+  GEnumValue *value;
+  gchar *name = NULL;
+
+  type_class = g_type_class_ref (ATK_TYPE_TEXT_ATTRIBUTE);
+  g_return_val_if_fail (G_IS_ENUM_CLASS (type_class), NULL);
+
+  value = g_enum_get_value (G_ENUM_CLASS (type_class), attr);
+
+  if (value)
+    {
+      name = value->value_nick;
+    }
+  else
+    {
+      if (extra_attributes)
+        {
+          gint n = attr;
+
+          n -= ATK_TEXT_ATTR_LAST_DEFINED + 1;
+
+          if (n < extra_attributes->len)
+
+            name = g_ptr_array_index (extra_attributes, n);
+        }
+    }
+  g_type_class_unref (type_class);
+  return name;
 }
 
 /**
- * atk_attribute_get_value:
+ * atk_text_attribute_for_name:
+ * @name: a string which is the (non-localized) name of an ATK text attribute.
+ *
+ * Get the #AtkTextAttribute type corresponding to a text attribute name.
+ *
+ * Returns: the #AtkTextAttribute enumerated type corresponding to the specified
+name,
+ *          or #ATK_TEXT_ATTRIBUTE_INVALID if no matching text attribute is found.
+ **/
+AtkTextAttribute
+atk_text_attribute_for_name (const gchar *name)
+{
+  GTypeClass *type_class;
+  GEnumValue *value;
+  AtkTextAttribute type = ATK_TEXT_ATTR_INVALID;
+
+  g_return_val_if_fail (name, ATK_TEXT_ATTR_INVALID);
+
+  type_class = g_type_class_ref (ATK_TYPE_TEXT_ATTRIBUTE);
+  g_return_val_if_fail (G_IS_ENUM_CLASS (type_class), ATK_TEXT_ATTR_INVALID);
+
+  value = g_enum_get_value_by_nick (G_ENUM_CLASS (type_class), name);
+
+  if (value)
+    {
+      type = value->value;
+    }
+  else
+    {
+      gint i;
+
+      if (extra_attributes)
+        {
+          for (i = 0; i < extra_attributes->len; i++)
+            {
+              gchar *extra_attribute = (gchar *)g_ptr_array_index (extra_attributes, i);
+
+              g_return_val_if_fail (extra_attribute, ATK_TEXT_ATTR_INVALID);
+
+              if (strcmp (name, extra_attribute) == 0)
+                {
+                  type = i + 1 + ATK_TEXT_ATTR_LAST_DEFINED;
+                  break;
+                }
+            }
+        }
+    }
+  g_type_class_unref (type_class);
+
+  return type;
+}
+
+
+/**
+ * atk_text_attribute_get_value:
  * @attr: The #AtkTextAttribute for which a value is required
  * @index: The index of the required value
  *
@@ -916,8 +1009,8 @@ atk_attribute_get_name (AtkTextAttribute attr)
  * NULL is returned if there are no values maintained for the attr value. 
  **/
 G_CONST_RETURN gchar*
-atk_attribute_get_value (AtkTextAttribute attr,
-                         gint             index)
+atk_text_attribute_get_value (AtkTextAttribute attr,
+                              gint             index)
 {
   switch (attr)
     {
